#! /usr/bin/perl -w
# Forfatter: Ulrik Dickow, udickow at gmail dot com, 20170722.
# Licens: Perl Artistic License, http://dev.perl.org/licenses/artistic.html
# Brug: $0 [FIL]...
# Formål: Omdan givet/givne spilresultater for en parturnering i html-form fra BridgeCentral
#   version 2 til en skifteplan i barometerform for turneringen.  Spilnumre i input ignoreres,
#   så man kan sagtens kombinere flere sektioner med genbrug af spilnumre til én samlet
#   skifteplan.  Angives ingen filer, læses fra standard-input.  Dermed kan man konvertere
#   direkte fra en webside ved kombination med fx curl ( https://curl.haxx.se/ ):
#
#     curl -s http://bridge.dk/2017/Klub7/Turneringer/SpilRes_S563_CG1.html | bc2-spilres-...
#
#   Outputformatet er Hollandsk ASC-standard: http://www.pjms.nl/movementfile_standard.html .
#   (Det kan konverteres til BridgeCentrals .lcd-format via programmet "manipulate", se link).
#   Hvis parrene i input er nummereret i bogstavgrupper (som i store festivalturneringer),
#   fx A1..A28,B1..B28, så omdøbes de i output til en sekvens af heltal, her 1..56.
#   Hvis parrene i visse spil har siddet omvendt, rettes dette tilbage i output.  Ved bogstaver
#   i parnumre mangler bogstavet i så fald for det aktuelle NS-par (som skulle sidde ØV);
#   men vi antager at det er lig bogstavet for ØV-parret (som skulle sidde NS) uden at checke
#   spillernes navne.  Den antagelse går i hvert fald godt i denne sølvturnering fra 2017:
#
#     http://bridge.dk/2017/Klub7/Turneringer/SpilRes_S503_CG1.html?ID=18365#spil31
#
#   Formålet er ikke at man skal spille efter den resulterende, ofte rodede, skifteplan.
#   I stedet er formålet at kunne regne på balance mv. for allerede afholdte turneringer uden
#   at skulle besvære turneringens arrangører med at fremfinde den gamle skifteplan.

use strict;

# GCD = Greatest Common Divisor, Euklids algoritme (Abelson, Sussman & Sussman (1985) side 44).
sub gcd2 { my($a,$b)=@_; $b == 0 ? $a : gcd2($b, $a % $b) }          # GCD for præcis 2 tal
sub gcd  { my($a,$b,@r)=@_; defined $b ? gcd(gcd2($a,$b), @r) : $a } # GCD for liste af tal

sub max2 { my($a,$b)=@_; $a > $b ? $a : $b }

sub bypairno { # Til sortering af originale parnumre, enten 1-2-3... eller A1-A2-...-B1-B2...
    "$a,$b" =~ /^([A-Z]?)(\d+),([A-Z]?)(\d+)/; # Matcher altid.
    return ($1 cmp $3 or $2 <=> $4); # A før B.  Hvis samme/intet bogstav, så bare numerisk.
}

sub same_seatings ($$) { # Sand hvis de to givne spil har samme NS/ØV-opgør, ellers falsk
    my ($h1, $h2) = @_;
    return 0 unless keys %$h1 == keys %$h2; # Falsk hvis ej samme antal opgør
    foreach (keys %$h1) {
	return 0 unless exists $h2->{$_} and $h2->{$_} eq $h1->{$_};
    }
    return 1;
}

sub renum_pairs {
    my ($h1, $h2) = @_; # Oversæt alle nøgler og værdier i h1 via oversættelse givet i h2
    my $renum = {};
    while (my ($key1, $val1) = each %$h1) {
	$renum->{$h2->{$key1}} = $h2->{$val1};
    }
    return $renum;
}

# Stak til opsamling af spil, ét element på stakken pr. spil i den rækkefølge de ses i input.
# Hvert element repræsenterer spillet med en hash: nøgler er NS-spillerne, værdier tilhørende ØV.
# Vi gemmer selvfølgelig kun parnumre, ikke spillernes navne.  "Oversidder" særbehandles ikke.
# De brugte par"numre" her er de originale i input, dvs. inkl. eventuelt forbogstav.  Først til
# sidst kan vi være sikre på at have set alle par og dermed kunne bestemme deres nye numre.
# Ved forkert led ("*" i input) ombyttes og gættes som beskrevet ovenfor.
my @boards;

my $board = {}; # Hashref til opsamling af det enkelte spil undervejs, inden det lægges på stak.

# Hash til oversættelse af parnumre fra input til output.  Undervejs blot 1 for alle som parnr;
# først til sidst kommer de rigtige i.
my %pairno;

my ($ns, $ew, $swap); # Originale parnumre under opsamling, plus plads til evt. ombytning af dem
my ($nt, $ntnow) = (0,0); # Antal borde et spil spilles ved (nt = maksimale, ntnow = nuværende)

# Vi laver primitiv tolkning linjevis af den ret grimme html fra BC 2.4.5.  Vil næppe du med BC3.
while (<>) {
    if (/<td NAME=.*?nbsp;(\*?)(([A-Z]?)\d+)/) { # Ny NS/ØV mødt, enten i samme eller nyt spil
	$ns = $2, next unless defined $ns;   # NS huskes blot.  Når ØV ses, gemmer vi parret.
	$ew = $2;                            # NS og ØV gemmes uden eventuelle stjerne på
	my ($ewstar, $ewletter) = ($1, $3);
	if ($ewstar) { # NS og ØV sad forkert.  Ombyt og gæt eventuelt manglende $ns-bogstav.
	    $swap = $ns; $ns = $ew;
	    $ew = ($ewletter && $swap =~ /^\d/ ? "$ewletter$swap" : $swap);
	}
	$board->{$ns} = $ew;
	$pairno{$ns} = $pairno{$ew} = 1;
	undef $ns;    # Så vi er klar til at opsamle nye NS/ØV for spillet
    } elsif (m,</table>, and %$board) {  # Vi har gemt mindst ét NS/ØV-par og spillet slut nu
	$nt = $ntnow if ($ntnow = keys %$board) > $nt;
	push @boards, $board;
	$board = {};   # Nulstil til at pege på nyt tomt spil, klar til nye NS/ØV-par
    }
}

my $np=0; # Ender med at være antal par i turneringen (= antal forskellige parnumre)
foreach my $pstr (sort bypairno keys %pairno) { # Oversæt til numeriske fra 1 til antal par
    $pairno{$pstr} = ++$np;
}

# Bestem antal runder.  I værste fald lig antal spil, men forhåbentlig kun en brøkdel af det.
# Bestemmes ved at finde antal spil i træk med samme NS-ØV-opgør.  Kan variere i fx visse
# serieturneringer, så find alle sådanne antal og afslutningsvis største fælles divisor.
my %nsame; # Nøgle k findes hvis sekvens med k ens NS-ØV-opgør fundet
for (my $isame=1, my $i=1; $i<@boards; $i++) {
    if (same_seatings $boards[$i-1], $boards[$i]) {
	$isame++;
    } else { # Gem forrige optælling og begynd ny
	$isame = $nsame{$isame} = 1;
    }
}
my $boards_per_round = gcd(keys %nsame);
my $nr = @boards / $boards_per_round;

# Print resulterende skifteplan: 1 linje header, $nr linjer med 1 linje pr. runde (= kortgruppe).
printf "%2d %2d %2d %2d 0\n", $np, $nt, $nr, $nr;
for my $r (1 .. $nr) {
    my $brd = renum_pairs($boards[($r-1) * $boards_per_round], \%pairno);
    my @ns = sort { max2($b, $brd->{$b}) <=> max2($a, $brd->{$a}) } keys %$brd;
    foreach (@ns) { printf "%2d-%2d %2d ", $_, $brd->{$_}, $r }
    for (1 .. $nt-@ns) { printf " 0- 0  0 " } # Hvis ikke alle spil spilles lige mange gange
    printf "\n";
}
